/*****************************************************************************/
/*
                                 SeSoLa.c


    THE GNU GENERAL PUBLIC LICENSE APPLIES DOUBLY TO ANYTHING TO DO WITH
                    AUTHENTICATION AND AUTHORIZATION!

    This package is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License, or any later
    version.

>   This package is distributed in the hope that it will be useful,
>   but WITHOUT ANY WARRANTY; without even the implied warranty of
>   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
>   GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


Named "SeSoLa" to avoid any confusion and/or conflict with OpenSSL routines.

Supports SSL v2 and v3 (TLSv1) protocols.

SSL I/O is implemented as a OpenSSL BIO_METHOD, named "sesola_method". It
provides NON-BLOCKING SSL input/output. All routines that are part of this
functionality are named "sesola_..." and are grouped towards the end of this
module.


OPENSSL
-------
With the retirement of Eric Young from OpenSource software and the creation of
the OpenSSL organisation this module will be based on this group's packages and
in particular the the VMS port provided by Richard Levitte.

  richard@levitte.org
  http://www.free.lp.se/openssl/

The initial adaption has been made using 0.9.2b-RL.


VERSION HISTORY
---------------
18-APR-99  MGD  improve error reporting during certificate loading
03-APR-99  MGD  support OpenSSL (with initial support for SSLeay 0.8 & 0.9),
                add WATCH callbacks
31-MAR-99  MGD  bugfix; report request from non-SSL port
20-JAN-99  MGD  report format refinenment
07-NOV-98  MGD  WATCH facility
24-OCT-98  MGD  per-service context (implies per-service certificate)
01-JUN-98  MGD  builds OK with SSLeay 0.9.0b
25-JAN-98  MGD  initial development for v5.0, SSLeay v0.8.1
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <string.h>

/* VMS related header files */
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "wasd.h"

#define WASD_MODULE "SESOLA"

/***************************************/
#ifdef SESOLA  /* secure sockets layer */
/***************************************/

/* OpenSSL header files */
#include "bio.h"
#include "buffer.h"
#include "crypto.h"
#include "err.h"
#include "httpd.h"
#include "ssl.h"
#include "X509.h"

/**********/
/* macros */
/**********/

/* i.e. versions 2 and 3 */
#define DEFAULT_SSL_PROTOCOL_VERSION 23

#define SESOLA_DEBUG_STREAM 0

#define SESOLA_UNDEFINED           0
#define SESOLA_ACCEPT              1
#define SESOLA_SHUTDOWN            2
#define SESOLA_SHUT                3
#define SESOLA_READ_DATA_BLOCK     4
#define SESOLA_READ_DATA_NOBLOCK   5
#define SESOLA_WRITE_DATA_BLOCK    6
#define SESOLA_WRITE_DATA_NOBLOCK  7

/**************************************/
/* macros to support SSLeay 0.8 & 0.9 */
/**************************************/

#ifndef OPENSSL_VERSION_NUMBER

#define SSL_library_init SSLeay_add_ssl_algorithms

#endif

/*******************/
/* data structures */
/*******************/

struct SeSoLaStruct
{
   /* OpenSSL structures */
   SSL  *SslPtr;
   BIO  *BioPtr;
   BIO  *BioSslPtr;
   SSL_CTX  *SeSoLaCtx;

   /* pointer back to the request owning this structure */
   struct RequestStruct  *rqptr;

   /*
       This "in-progress" flag is to provide an additional indication of
       when a OpenSSL routine has used non-blocking I/O. It is only valid
       within the single AST-delivery context (i.e. to be checked immediately
       after using an OpenSSL function, before returning out of the current
       AST delivery context), because as soon as the I/O AST is delivered
       the approriate flags will be reset.
   */
   boolean  IoInProgress;

   int  /* */
        AcceptCount,
        /* number of characters currently available in read buffer */
        ReadCount,
        /* size of read data buffer (for AST routine) */
        ReadDataSize,
        /* flag indicating purpose and/or type of network read */
        ReadType,
        /* number of characters actually written to network */
        WriteCount,
        /* number of characters to be written to network */
        WriteDataLength,
        /* flag indicating purpose and/or type of network write */
        WriteType;

  char  /* during SSL connect accept points to start of buffer */
        *AcceptReadDataPtr,
        /* pointer to read data buffer (for AST routine use) */
        *ReadDataPtr,
        /* pointer to write data buffer (for AST routine use) */
        *WriteDataPtr;

   /* SSL post-processing, stores the pointer to the AST routine */
   void  *AstFunctionPtr;
};

/******************/
/* global storage */
/******************/

boolean  ProtocolHttpsAvailable,
         ProtocolHttpsConfigured;

int SeSoLaProtocolVersion = DEFAULT_SSL_PROTOCOL_VERSION;

char  HttpdSesola [] = " SSL";

char  ErrorSslPort [] = "This is an SSL (&quot;https:&quot;) port.",
      ErrorSslReportService [] =
        "This report is only available via an SSL service!";

/********************/
/* external storage */
/********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

extern int  AdminWatchEnabled;
extern int  NetReadBufferSize;
extern int  OutputBufferSize;
extern char  ErrorSanityCheck[];
extern char  ServerHostPort[];
extern char  Utility[];
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;
extern struct MsgStruct  Msgs;
extern struct ServiceStruct  *ServiceListHead;

/**********************************/
/* BIO method function prototypes */
/**********************************/

int sesola_write (BIO*, char*, int);
int sesola_write_ast (struct RequestStruct*);
int sesola_read (BIO*, char*, int);
int sesola_read_ast (struct RequestStruct*);
int sesola_puts (BIO*, char*);
int sesola_gets (BIO*, char*, int);
long sesola_ctrl (BIO*, int, long, char*);
int sesola_new (BIO*);
int sesola_free (BIO*);

int SesolaWatchBioCallback (BIO*, int, char*, int, long, long);
SesolaWatchInfoCallback (SSL*, int, int);

BIO_METHOD *BIO_s_sesola();

BIO_METHOD sesola_method =
{
   BIO_TYPE_FD,
   "WASD SESOLA",
   sesola_write,
   sesola_read,
   sesola_puts,
   sesola_gets,
   sesola_ctrl,
   sesola_new,
   sesola_free,
};

/*****************************************************************************/
/*
*/

SeSoLaInit ()

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaInit()\n");

   if (SeSoLaProtocolVersion < 0) return;
   if (!SeSoLaProtocolVersion)
      SeSoLaProtocolVersion = DEFAULT_SSL_PROTOCOL_VERSION;
   if (SeSoLaProtocolVersion != 2 &&
       SeSoLaProtocolVersion != 3 &&
       SeSoLaProtocolVersion != 23)
   {
      fprintf (stdout, "%%%s-E-SSL, unknown SSL protocol version\n \\%d\\\n",
               Utility, SeSoLaProtocolVersion);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }

   SSL_load_error_strings ();

   SSL_library_init (); 

   ProtocolHttpsAvailable = true;

   fprintf (stdout, "%%%s-I-SSL, %s\n",
            Utility, SSLeay_version(SSLEAY_VERSION));
}

/*****************************************************************************/
/*
Scan through the list of services initializeing an SSL context for each SSL
service.
*/

SeSoLaInitServiceList ()

{
   register char  *cptr, *sptr, *zptr;
   register struct ServiceStruct  *svptr;

   int  value;
   char  *DefaultSeSoLaCertPtr;
   SSL_CTX  *SeSoLaCtx;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaInitServiceList()\n");

   DefaultSeSoLaCertPtr = getenv ("HTTPD$SSL_CERT");

   for (svptr = ServiceListHead; svptr != NULL; svptr = svptr->NextPtr)
   {
      if (svptr->RequestScheme != SCHEME_HTTPS) continue;

      if (!svptr->SeSoLaCert[0])
      {
         zptr = (sptr = svptr->SeSoLaCert) + sizeof(svptr->SeSoLaCert);
         for (cptr = DefaultSeSoLaCertPtr;
              *cptr && sptr < zptr;
              *sptr++ = *cptr++);
         if (sptr >= zptr) ErrorExitVmsStatus (0, "string overflow", FI_LI);
         *sptr = '\0';
      }
      fprintf (stdout, "%%%s-I-SSL, %s %s\n",
               Utility, svptr->ServerHostPort, svptr->SeSoLaCert);

      if (SeSoLaProtocolVersion == 2)
         SeSoLaCtx = SSL_CTX_new (SSLv2_server_method());
      else
      if (SeSoLaProtocolVersion == 3)
         SeSoLaCtx = SSL_CTX_new (SSLv3_server_method());
      else
      if (SeSoLaProtocolVersion == 23)
         SeSoLaCtx = SSL_CTX_new (SSLv23_server_method());
      if (SeSoLaCtx == NULL)
         ErrorExitVmsStatus (0, "invalid SSL CTX", FI_LI);

      /* turn on SYSPRV to ensure access to the certificate file */
      EnableSysPrv ();

      value = SSL_CTX_use_certificate_file (SeSoLaCtx,
                                            svptr->SeSoLaCert,
                                            SSL_FILETYPE_PEM);
      if (!value)
      {
         SeSoLaPrintOpenSslErrorList ();
         ErrorExitVmsStatus (0, "SSL server certificate file", FI_LI);
      }

      value = SSL_CTX_use_RSAPrivateKey_file (SeSoLaCtx,
                                              svptr->SeSoLaCert,
                                              SSL_FILETYPE_PEM);
      if (!value)
      {
         SeSoLaPrintOpenSslErrorList ();
         ErrorExitVmsStatus (0, "SSL using RSA private key file", FI_LI);
      }

      value = SSL_CTX_check_private_key (SeSoLaCtx);

      if (!value)
      {
         SeSoLaPrintOpenSslErrorList ();
         ErrorExitVmsStatus (0, "SSL private key check", FI_LI);
      }

      DisableSysPrv ();

      /* 
         Seems to allow for a swag of browser bugs (see [.INCLUDE]SSL.H).
         Problems were only encountered when moving to OpenSSl v0.9.2B-RL,
         (perhaps it was on by default in earlier versions).
      */
      SSL_CTX_set_options (SeSoLaCtx,SSL_OP_ALL);

      (SSL_CTX*)svptr->SeSoLaCtx = SeSoLaCtx;
   }
}

/*****************************************************************************/
/*
Print out the OpenSSL error list.  For reporting errors associated with
certificate loading.  Based on [.CRYPTO.ERR]ERR_PRN.C.
*/

SeSoLaPrintOpenSslErrorList ()

{
   int  LineNumber,
        Flags;
   unsigned long  ErrorNumber;
   char  *FileNamePtr,
         *DataStringPtr;
   char  Buffer [200];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaPrintOpenSslErrorList()\n");

   while (ErrorNumber =
          ERR_get_error_line_data (&FileNamePtr, &LineNumber,
                                   &DataStringPtr, &Flags))
   {
      fprintf (stdout, "%%%s-E-SSL, \"%s\" (%s %d)%s%s%s\n",
               Utility,
               ERR_error_string (ErrorNumber, Buffer),
               FileNamePtr, LineNumber,
               ((Flags & ERR_TXT_STRING) ? " \"" : ""),
               ((Flags & ERR_TXT_STRING) ? DataStringPtr : ""),
               ((Flags & ERR_TXT_STRING) ? "\"" : ""));
   }
}

/*****************************************************************************/
/*
Begin an SSL transaction for a request. Create the SESOLA structure used to
store HTTPd SSL-related used during the transaction, initialize the OpenSSL
structures required, then begin the OpenSSL accept functionality.
*/

SeSoLaBegin (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaBegin() %d\n", rqptr);

   /* note that this IS NOT HEAP memory, and must be explicitly VmFree()d */
   rqptr->SeSoLaPtr = sesolaptr =
      (struct SeSoLaStruct*)VmGet (sizeof(struct SeSoLaStruct));

   sesolaptr->SeSoLaCtx = (SSL_CTX*)rqptr->ServicePtr->SeSoLaCtx;

   /* set the SeSoLa structure request pointer */
   sesolaptr->rqptr = rqptr;

   if (Debug) fprintf (stdout, "SSL_new()\n");
   sesolaptr->SslPtr = SSL_new (sesolaptr->SeSoLaCtx);
   if (sesolaptr->SslPtr == NULL)
   {
      fprintf (stdout, "%%%s-W-SSL, SSL_new(sesolaptr->SeSoLaCtx) failed\n",
               Utility);
      SeSoLaFree (rqptr);

      if (rqptr->AdminWatchItem &&
          (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "BEGIN failed");

      return;
   }

   if (Debug) fprintf (stdout, "BIO_new()\n");
   sesolaptr->BioPtr = BIO_new (BIO_s_sesola());
   if (sesolaptr->BioPtr == NULL)
   {
      fprintf (stdout, "%%%s-W-SSL, BIO_new(BIO_s_sesola()) failed\n",
               Utility);
      SeSoLaFree (rqptr);

      if (rqptr->AdminWatchItem &&
          (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "BEGIN failed");

      return;
   }
   /* set the "BIO_s_sesola" pointer to the SESOLA structure */
   sesolaptr->BioPtr->ptr = sesolaptr;

   /* set up the SSL filter */
   if (Debug) fprintf (stdout, "BIO_new()\n");
   sesolaptr->BioSslPtr = BIO_new (BIO_f_ssl());
   if (sesolaptr->BioSslPtr == NULL)
   {
      fprintf (stdout, "%%%s-W-SSL, BIO_new(BIO_f_ssl()) failed\n", Utility);
      SeSoLaFree (rqptr);

      if (rqptr->AdminWatchItem &&
          (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "BEGIN failed");

      return;
   }

   SSL_set_accept_state (sesolaptr->SslPtr);
   SSL_set_bio (sesolaptr->SslPtr, sesolaptr->BioPtr, sesolaptr->BioPtr);

   if (rqptr->AdminWatchItem &&
       (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
   {
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "BEGIN");
      /* set the request pointer as the first item of SSL user data */
      SSL_set_ex_data (sesolaptr->SslPtr, 0, rqptr);
      /* set for SSL information and verification callback */
      SSL_CTX_set_info_callback (sesolaptr->SslPtr, &SesolaWatchInfoCallback);
      /* set for BIO callback with the request pointer as it's argurment */
      BIO_set_callback (sesolaptr->BioPtr, &SesolaWatchBioCallback);
      BIO_set_callback_arg (sesolaptr->BioPtr, rqptr);
   }
}

/*****************************************************************************/
/*
Shutdown the SSL session with the client (via an SSL alert that we don't wait
around for a response on!)  Return true if RequestEnd() can continue
the request run-down, false if has to abort and wait for some more to happen!
*/

boolean SeSoLaEnd (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaEnd() %d\n", rqptr);

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   if (sesolaptr->ReadType == SESOLA_SHUT &&
       sesolaptr->WriteType == SESOLA_SHUT)
   {
      /* final call from SysDclAst() below invoked RequestEnd() */

      if (rqptr->AdminWatchItem &&
          (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "END");

      return (true);
   }

   if (sesolaptr->ReadType == SESOLA_SHUTDOWN &&
       sesolaptr->WriteType == SESOLA_SHUTDOWN)
   {
      /* AST call from sesola_write() in SSL_shutdown() below */
      sesolaptr->ReadType = sesolaptr->WriteType = SESOLA_SHUT;
      SSL_set_shutdown (sesolaptr->SslPtr,
                        SSL_SENT_SHUTDOWN | SSL_RECEIVED_SHUTDOWN);
      /* just in case there's some non-block I/O I don't count on! */
      if (sesolaptr->IoInProgress) return (false);

      /* now, run down the request again */
      SysDclAst (&RequestEnd, rqptr);
      return (false);
   }

   sesolaptr->ReadType = sesolaptr->WriteType = SESOLA_SHUTDOWN;

   /* send the SSL close alert to the client */
   value = SSL_shutdown (sesolaptr->SslPtr);
   if (Debug) fprintf (stdout, "SSL_shutdown() %d\n", value);

   /* return if the shutdown is not yet complete (write underway) */
   if (value <= 0 || sesolaptr->IoInProgress) return (false);

   /* looks like we're all shutdown, just continue RequestEnd() */

   if (rqptr->AdminWatchItem &&
       (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA, "END");

   sesolaptr->ReadType = sesolaptr->WriteType = SESOLA_SHUT;
   return (true);
}

/*****************************************************************************/
/*
Free the OpenSSL structures outside of any use of them (early mistake)!
*/

boolean SeSoLaFree (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaFree() %d\n", rqptr);

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   if (sesolaptr->BioPtr != NULL) BIO_free (sesolaptr->BioPtr);

   if (sesolaptr->SslPtr != NULL) SSL_free (sesolaptr->SslPtr);

   /* request has finished with the SESOLA structure */
   VmFree (sesolaptr);
   rqptr->SeSoLaPtr = NULL;
}

/*****************************************************************************/
/*
This establishes the SSL transaction by providing the server "hello",
certificate and key exchange, etc. Due to the non-blocking I/O used by WASD
this function will be called multiple times to complete the OpenSSL accept. 
*/

SeSoLaAccept (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   int  status,
        value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaAccept() %d\n", rqptr);

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   sesolaptr->ReadType = sesolaptr->WriteType = SESOLA_ACCEPT;

   if (++sesolaptr->AcceptCount == 2)
   {
      /* just finished the first lot of non-blocking I/O */
      if (!memcmp (sesolaptr->AcceptReadDataPtr, "GET ", 4) ||
          !memcmp (sesolaptr->AcceptReadDataPtr, "HEAD ", 5) ||
          !memcmp (sesolaptr->AcceptReadDataPtr, "POST ", 5) ||
          !memcmp (sesolaptr->AcceptReadDataPtr, "PUT ", 4) ||
          !memcmp (sesolaptr->AcceptReadDataPtr, "DELETE ", 7))
      {
         /* looks like it's a standard HTTP request on the wrong port! */
         rqptr->ResponseStatusCode = 400;
         ErrorGeneral (rqptr, ErrorSslPort, FI_LI);
         SysDclAst (&RequestEnd, rqptr);
         SeSoLaFree (rqptr);
         return;
      }
   }

   value = SSL_accept (sesolaptr->SslPtr);
   if (Debug) fprintf (stdout, "SSL_accept() %d\n", value);

#ifdef DBUG
   if (Debug) ERR_print_errors_fp (stdout);
#endif

   switch (SSL_get_error (sesolaptr->SslPtr, value))
   {
      case SSL_ERROR_NONE :
         break;
      case SSL_ERROR_WANT_WRITE :
      case SSL_ERROR_WANT_READ :
      case SSL_ERROR_WANT_X509_LOOKUP :
         return;
      case SSL_ERROR_SYSCALL :
      case SSL_ERROR_SSL :
      case SSL_ERROR_ZERO_RETURN :
      default:
         if (value == -1 || sesolaptr->IoInProgress)
         {
            NetShutdownSocket (rqptr, 0, true);
            return;
         }
         SysDclAst (&RequestEnd, rqptr);
         return;
   }

#ifdef DBUG
   if (Debug) SeSoLaDebugSession (sesolaptr);
#endif

   if (rqptr->AdminWatchItem &&
       (AdminWatchEnabled & ADMIN_WATCH_SESOLA))
      SesolaWatchSession (rqptr);

   rqptr->NetReadBufferPtr = VmGetHeap (rqptr, NetReadBufferSize);
   rqptr->NetReadBufferSize = NetReadBufferSize;

   NetRead (rqptr, &RequestGet,
            rqptr->NetReadBufferPtr, rqptr->NetReadBufferSize);
}

/*****************************************************************************/
/*
Read application data. This is the equivalent of the general NetRead() function
(and is indeed called from it) in that it reads data from the client via the
network. This function uses SSL_read() to read a stream of encrypted data from
the network then provide it decrypted in the supplied buffer.  This function
provides BLOCKING (non- AST) and NON-BLOCKING I/O conforming to the
functionality provided by NetRead(). The network read I/O status block status
and count fields are valid for the high-level calling routine.
*/

SeSoLaRead
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataSize
)
{
   register struct SeSoLaStruct  *sesolaptr;

   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaRead() %d\n", rqptr);

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   if (Debug) fprintf (stdout, "sesolaptr->ReadCount: %d\n", sesolaptr->ReadCount);

   if (!sesolaptr->ReadCount)
   {
      if (AstFunctionPtr == NULL)
         sesolaptr->ReadType = SESOLA_READ_DATA_BLOCK;
      else
      {
         sesolaptr->ReadType = SESOLA_READ_DATA_NOBLOCK;
         sesolaptr->AstFunctionPtr = AstFunctionPtr;
      }
   }

   value = SSL_read (sesolaptr->SslPtr,
                     sesolaptr->ReadDataPtr = DataPtr,
                     sesolaptr->ReadDataSize = DataSize);
   if (Debug) fprintf (stdout, "SSL_read() %d\n", value);

   /* return if non-blocking I/O is in progress */
   if (value == -1 || sesolaptr->IoInProgress) return (SS$_NORMAL);

   /* fudge the I/O status block count to reflect the application data count */
   rqptr->NetReadIOsb.Count = value;

   if (AstFunctionPtr == NULL) return (rqptr->NetReadIOsb.Status);

   SysDclAst (AstFunctionPtr, rqptr);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Write application data. This is the equivalent of the general NetWrite()
function (and is called from it) in that it writes data to the client via the
network. This function uses SSL_write() to write non-encrypted application data
to the network in encrypted form. This function provides BLOCKING (non-AST)
and NON-BLOCKING I/O conforming to the functionality provided by NetWrite().
The network write I/O status block status and count fields are valid for the
high-level calling routine.
*/

int SeSoLaWrite
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataLength
)
{
   register struct SeSoLaStruct  *sesolaptr;

   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SeSoLaWrite() %d %d %d\n", rqptr, DataPtr, DataLength);
   /** fprintf (stdout, "|%*.*s|\n", DataLength, DataLength, DataPtr); **/

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   if (Debug) fprintf (stdout, "sesolaptr->WriteCount: %d\n", sesolaptr->WriteCount);

   if (!sesolaptr->WriteCount)
   {
      if (AstFunctionPtr == NULL)
         sesolaptr->WriteType = SESOLA_WRITE_DATA_BLOCK;
      else
      {
         sesolaptr->WriteType = SESOLA_WRITE_DATA_NOBLOCK;
         sesolaptr->AstFunctionPtr = AstFunctionPtr;
      }
   }

   value = SSL_write (sesolaptr->SslPtr,
                      sesolaptr->WriteDataPtr = DataPtr,
                      sesolaptr->WriteDataLength = DataLength);
   if (Debug) fprintf (stdout, "SSL_write() %d\n", value);

   /* return if non-blocking I/O is in progress */
   if (value == -1 || sesolaptr->IoInProgress) return (SS$_NORMAL);

   if (AstFunctionPtr == NULL) return (rqptr->NetReadIOsb.Status);

   SysDclAst (AstFunctionPtr, rqptr);

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
Provides OpenSSL status information at various states of SSL processing via the
WATCH facility.
*/

SesolaWatchInfoCallback
(
SSL *SslPtr,
int WhereExactly,
int Value
)
{
   int  WhereGenerally;
   char  *cptr;
   struct  RequestStruct  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SesolaWatchInfoCallback()\n");

   rqptr = (struct RequestStruct*)SSL_get_ex_data (SslPtr, 0);

   WhereGenerally = WhereExactly & ~SSL_ST_MASK;

   if (WhereGenerally & SSL_ST_CONNECT)
      cptr = "SSL_CONNECT";
   else
   if (WhereGenerally & SSL_ST_ACCEPT)
      cptr = "SSL_ACCEPT";
   else
      cptr = "UNDEFINED";

   if (WhereExactly & SSL_CB_LOOP)
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                      "!AZ !AZ",
                      cptr, SSL_state_string_long(SslPtr));
   else
   if (WhereExactly & SSL_CB_ALERT)
   {
      cptr = (WhereExactly & SSL_CB_READ) ? "read" : "write";
      AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                      "!AZ SSL3 alert !AZ:!AZ",
                      cptr, SSL_alert_type_string_long(Value),
                      SSL_alert_desc_string_long(Value));
   }
   else
   if (WhereExactly & SSL_CB_EXIT)
   {
      if (Value == 0)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "!AZ SSL3 alert failed in !AZ",
                         cptr, SSL_state_string_long(SslPtr));
      else
      if (Value < 0)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "!AZ error in !AZ",
                         cptr, SSL_state_string_long(SslPtr));
   }
}

/*****************************************************************************/
/*
Provides OpenSSL Input/Output information each time an SSL read or write occurs
via the WATCH facility.
*/

int SesolaWatchBioCallback
(
BIO *bioptr,
int cmd,
char *argp,
int argi,
long argl,
long retval
)
{
   struct  RequestStruct  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SesolaWatchBioCallback()\n");

   rqptr = (struct RequestStruct*) BIO_get_callback_arg (bioptr);

   if (cmd == (BIO_CB_READ | BIO_CB_RETURN))
   {
      if (retval == -1)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "READ !UL bytes non-blocking", argi);
      else
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "READ !UL bytes done", retval);
      return (retval);
   }
   else
   if (cmd == (BIO_CB_WRITE | BIO_CB_RETURN))
   {
      if (retval == -1)
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "WRITE !UL bytes non-blocking", argi);
      else
         AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                         "WRITE !UL bytes done", retval);
      return (retval);
   }

   return (retval);
}

/*****************************************************************************/
/*
Provide WATCH information after final SSL session establishment show version
and cipher in use.  Code based in part on OpenSSL SSL_session_print().
*/

SeSoLaWatchSession (struct RequestStruct *rqptr)

{
   char  *CipherPtr,
         *VersionPtr;
   char  CipherNumber [64];
   SSL  *SslPtr;
   SSL_SESSION  *SessionPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSoLaWatchSession()\n");

   SslPtr = rqptr->SeSoLaPtr->SslPtr;

   SessionPtr = SSL_get_session (SslPtr);

   if (SessionPtr->ssl_version == SSL2_VERSION)
      VersionPtr = "SSLv2";
   else
   if (SessionPtr->ssl_version == SSL3_VERSION)
      VersionPtr = "SSLv3";
   else
   if (SessionPtr->ssl_version == TLS1_VERSION)
      VersionPtr = "TLSv1";
   else
      VersionPtr = "unknown";

   if (SessionPtr->cipher == NULL)
   {
      if (((SessionPtr->cipher_id) & 0xff000000) == 0x02000000)
         sprintf (CipherPtr = CipherNumber, "%06lX",
                  SessionPtr->cipher_id & 0xffffff);
      else
         sprintf (CipherPtr = CipherNumber, "%04lX",
                  SessionPtr->cipher_id & 0xffff);
   }
   else
      CipherPtr = SessionPtr->cipher->name;

   AdminWatchThis (rqptr, FI_LI, ADMIN_WATCH_SESOLA,
                   "SESSION !AZ protocol:!AZ cipher:!AZ",
                   (SslPtr->hit ? "(cached)" : "(new)"),
                   VersionPtr, CipherPtr);
}

/*****************************************************************************/
/*
*/

SeSoLaReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ ... SSL Report</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1><NOBR>HTTPd !AZ</NOBR></H1>\n\
<H2>SSL Report, !AZ</H2>\n\
!AZ\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=2>\n\
<TR><TH COLSPAN=2>!AZ</TH></TR>\n\
<TR><TH>General</TH><TH>Supported Ciphers</TH></TR>\n\
<TR><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right><U>Service Accepts</U></TH></TR>\n\
<TR><TH ALIGN=right>Total:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Completed:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right><U>Session cache</U></TH></TR>\n\
<TR><TH ALIGN=right>Hits:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Misses:</TH><TD>!UL</TD></TR>\n\
<TR><TH ALIGN=right>Timeouts:</TH><TD>!UL</TD></TR>\n\
</TABLE>\n\
</TD><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH></TH>\
<TH ALIGN=left><U>Version</U></TH>\
<TH ALIGN=left><U>Name</U></TH>\
</TR>\n");

   static $DESCRIPTOR (CiphersFaoDsc,
"<TR><TH ALIGN=right>!UL. &nbsp;</TH>\
<TD><TT>!AZ</TT> &nbsp;</TD>\
<TD><TT>!AZ</TT> &nbsp;</TD>\
</TR>\n");

   static $DESCRIPTOR (CurrentSession1FaoDsc,
"</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=3 CELLSPACING=0 BORDER=2>\n\
<TR><TH>Current Session</TH></TR>\n\
<TR><TD VALIGN=top>\n\
<TABLE CELLPADDING=0 CELLSPACING=5 BORDER=0>\n\
<TR><TH ALIGN=right>Cache Hit:</TH><TD>!AZ</TD></TR>\n\
<TR><TH ALIGN=right>Version:</TH><TD><TT>!AZ</TD></TT></TR>\n\
<TR><TH ALIGN=right VALIGN=top ROWSPAN=!UL>Shared&nbsp;Ciphers:</TH>");

   static $DESCRIPTOR (SharedCipherFaoDsc,
"!AZ<TD><B>!UL.</B> &nbsp;<TT>!AZ</TT></TD></TR>\n");

   static $DESCRIPTOR (CurrentSession2FaoDsc,
"<TR><TH ALIGN=right>Session Cipher:</TH><TD><TT>!AZ</TT></TD></TR>\n\
<TR><TH ALIGN=right>Session-ID:</TH><TD><TT>!#(2XB)</TT></TD></TR>\n\
<TR><TH ALIGN=right>Master-Key:</TH><TD><TT>!#(2XB)</TT></TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;
   register char  *cptr, *sptr, *zptr;
   register struct SeSoLaStruct  *sesolaptr;

   int  status,
        Count,
        Total;
   unsigned short  Length;
   unsigned long  BinaryTime [2],
                  /* WARNING: really does need to be fairly large!! */
                  FaoVector [256],
                  ResultTime [2];
   char  Buffer [2048],
         SharedCiphers [256];
   $DESCRIPTOR (BufferDsc, Buffer);
   STACK  *StackPtr;
   SSL_CIPHER  *CipherPtr;
   SSL_SESSION  *SessionPtr;
   X509  *PeerPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SeSolaReport()\n");

   if (rqptr->RequestScheme != SCHEME_HTTPS)
   {
      ErrorGeneral (rqptr, ErrorSslReportService, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   HTTP_HEADER_200_HTML (rqptr);

   vecptr = FaoVector;

   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostPort;
   *vecptr++ = ServerHostPort;
   *vecptr++ = rqptr->ServicePtr->ServerHostPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   *vecptr++ = SSLeay_version (SSLEAY_VERSION);
   *vecptr++ = SSL_CTX_sess_accept (sesolaptr->SeSoLaCtx);
   *vecptr++ = SSL_CTX_sess_accept_good (sesolaptr->SeSoLaCtx);
   *vecptr++ = SSL_CTX_sess_hits (sesolaptr->SeSoLaCtx);
   *vecptr++ = SSL_CTX_sess_misses (sesolaptr->SeSoLaCtx);
   *vecptr++ = SSL_CTX_sess_timeouts (sesolaptr->SeSoLaCtx);

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }

   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   if (rqptr->RequestScheme != SCHEME_HTTPS)
   {
      NetWriteBuffered (rqptr, NextTaskFunction, "</BODY>\n</HTML>\n", 16);
      return;
   }

   /*********************/
   /* supported ciphers */
   /*********************/

   StackPtr = SSL_get_ciphers (sesolaptr->SslPtr);
   Total = sk_num (StackPtr);
   for (Count = 0; Count < Total; Count++)
   {
      CipherPtr = (SSL_CIPHER *)sk_value (StackPtr, Count);

      vecptr = FaoVector;
      *vecptr++ = Count + 1;
      *vecptr++ = SSL_CIPHER_get_version (CipherPtr);
      *vecptr++ = SSL_CIPHER_get_name (CipherPtr);

      status = sys$faol (&CiphersFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, FI_LI);
         SysDclAst (NextTaskFunction, rqptr);
         return;
      }
      Buffer[Length] = '\0';
      NetWriteBuffered (rqptr, 0, Buffer, Length);
   }

   /*******************/
   /* current session */
   /*******************/

   vecptr = FaoVector;

   if (sesolaptr->SslPtr->hit)
      *vecptr++ = "yes";
   else
      *vecptr++ = "no";

   CipherPtr = SSL_get_current_cipher (sesolaptr->SslPtr);

   /* can't really be null or we would not be here! */
   if (CipherPtr != NULL)
      *vecptr++ = SSL_CIPHER_get_version (CipherPtr);

   cptr = sptr = SSL_get_shared_ciphers (sesolaptr->SslPtr,
                    SharedCiphers, sizeof(SharedCiphers));
   Count = 0;
   /* there has to be at least one or we would not be here! */
   if (cptr != NULL)
   {
      while (*cptr)
      {
         while (*cptr && *cptr != ':') cptr++;
         if (*cptr) cptr++;
         Count++;
      }
   }
   *vecptr++ = Count;

   status = sys$faol (&CurrentSession1FaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, 0, Buffer, Length);

   /* has to be at least one or we would not be here! */
   cptr = sptr;
   if (cptr != NULL)
   {
      Count = 0;
      while (*cptr)
      {
         vecptr = FaoVector;

         sptr = cptr;
         while (*cptr && *cptr != ':') cptr++;
         if (*cptr) *cptr++ = '\0';

         if (Count++)
            *vecptr++ = "<TR>";
         else
            *vecptr++ = "";

         *vecptr++ = Count;
         *vecptr++ = sptr;

         status = sys$faol (&SharedCipherFaoDsc, &Length, &BufferDsc,
                            &FaoVector);
         if (VMSnok (status) || status == SS$_BUFFEROVF)
         {
            rqptr->ErrorTextPtr = "sys$faol()";
            ErrorVmsStatus (rqptr, status, FI_LI);
            SysDclAst (NextTaskFunction, rqptr);
            return;
         }
         Buffer[Length] = '\0';
         NetWriteBuffered (rqptr, 0, Buffer, Length);
      }
   }

   vecptr = FaoVector;

   /* can't really be null or we would not be here! */
   if (CipherPtr != NULL)
      *vecptr++ = SSL_CIPHER_get_name (CipherPtr);

   /*******************************************/
   /* eqivalent to OpenSSL SSL_session_print() */
   /*******************************************/

   SessionPtr = SSL_get_session (sesolaptr->SslPtr);

   *vecptr++ = SessionPtr->session_id_length;
   for (Count = 0; Count < (int)SessionPtr->session_id_length; Count++)
      *vecptr++ = SessionPtr->session_id[Count];

   *vecptr++ = SessionPtr->master_key_length;
   for (Count = 0; Count < (int)SessionPtr->master_key_length; Count++)
      *vecptr++ = SessionPtr->master_key[Count];

   status = sys$faol (&CurrentSession2FaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, FI_LI);
      SysDclAst (NextTaskFunction, rqptr);
      return;
   }
   Buffer[Length] = '\0';
   NetWriteBuffered (rqptr, NextTaskFunction, Buffer, Length);
}

/*****************************************************************************/
/*
*/

#ifdef DBUG

SeSoLaDebugStream
(
char *DataPtr,
int DataLength
)
{
   /*********/
   /* begin */
   /*********/

   fprintf (stdout, "SeSoLaDebugStream() %d %d\n", DataPtr, DataLength);

   if (!SESOLA_DEBUG_STREAM) return;

   while (DataLength-- > 0)
      fprintf (stdout, "%02.02x ", (unsigned char)*DataPtr++);
   fprintf (stdout, "\n");
}

#endif

/*****************************************************************************/
/*
*/

#ifdef DBUG

SeSoLaDebugSession (struct SeSoLaStruct *sesolaptr)

{
   static BIO *bio_stdout = NULL;

   int  Count,
        Total;
   char  *cptr;
   char  Buffer [1024];
   STACK  *StackPtr;
   SSL_CIPHER  *CipherPtr;
   X509  *PeerPtr;

   /*********/
   /* begin */
   /*********/

   fprintf (stdout, "SeSoLaDebugSession()\n");

   if (bio_stdout == NULL)
      if ((bio_stdout = BIO_new (BIO_s_file())) != NULL)
         BIO_set_fp (bio_stdout, stdout, BIO_CLOSE);

   BIO_printf (bio_stdout, "----------\n");

   BIO_puts (bio_stdout, SSLeay_version (SSLEAY_VERSION));
   BIO_puts (bio_stdout, "\n");

   BIO_printf (bio_stdout, "Statistics:\n");
   BIO_printf (bio_stdout, "%4d client connects (SSL_connect())\n",
               SSL_CTX_sess_connect (sesolaptr->SeSoLaCtx));
   BIO_printf (bio_stdout, "%4d client connects that finished\n",
               SSL_CTX_sess_connect_good (sesolaptr->SeSoLaCtx));
   BIO_printf (bio_stdout, "%4d server accepts (SSL_accept())\n",
               SSL_CTX_sess_accept (sesolaptr->SeSoLaCtx));
   BIO_printf (bio_stdout, "%4d server accepts that finished\n",
               SSL_CTX_sess_accept_good (sesolaptr->SeSoLaCtx));
   BIO_printf (bio_stdout, "%4d session cache hits\n",
               SSL_CTX_sess_hits (sesolaptr->SeSoLaCtx));
   BIO_printf (bio_stdout, "%4d session cache misses\n",
               SSL_CTX_sess_misses (sesolaptr->SeSoLaCtx));
   BIO_printf (bio_stdout, "%4d session cache timeouts\n",
               SSL_CTX_sess_timeouts (sesolaptr->SeSoLaCtx));
   BIO_printf (bio_stdout, "%4d callback cache hits\n",
               SSL_CTX_sess_cb_hits (sesolaptr->SeSoLaCtx));
   BIO_printf (bio_stdout, "\n");

   StackPtr = SSL_get_ciphers (sesolaptr->SslPtr);
   Total = sk_num (StackPtr);
   BIO_printf (bio_stdout, "%d ciphers supported:\n", Total);
   for (Count = 0; Count < Total; Count++)
   {
      CipherPtr = (SSL_CIPHER *)sk_value (StackPtr, Count);
      BIO_printf (bio_stdout, "%s:%-25s",
                  SSL_CIPHER_get_version (CipherPtr),
                  SSL_CIPHER_get_name (CipherPtr));
      if ((((Count+1)%3) == 0) && (Count+1 != Total))
         BIO_puts (bio_stdout, "\n");
   }
   BIO_puts (bio_stdout, "\n");
   cptr = SSL_get_shared_ciphers (sesolaptr->SslPtr, Buffer, sizeof(Buffer));
   if (cptr != NULL)
   {
      BIO_printf (bio_stdout,
         "\nCiphers common between both SSL end points:\n");
      Total = Count = 0;
      while (*cptr)
      {
         if (*cptr == ':')
         {
            BIO_write (bio_stdout, "               ", 15-Count);
            Total++;
            Count=0;
            BIO_write (bio_stdout, ((Total%3) ? " " : "\n"), 1);
         }
         else
         {
            BIO_write (bio_stdout, cptr, 1);
            Count++;
         }
         cptr++;
      }
      BIO_puts (bio_stdout, "\n");
   }
   BIO_printf (bio_stdout, ((sesolaptr->SslPtr->hit) ? "\nReused, " : "\nNew, "));
   CipherPtr = SSL_get_current_cipher (sesolaptr->SslPtr);
   if (CipherPtr == NULL)
      BIO_printf (bio_stdout, "(none current)\n");
   else
   {
      BIO_printf (bio_stdout, "%s, Cipher is %s\n",
                  SSL_CIPHER_get_version (CipherPtr),
                  SSL_CIPHER_get_name (CipherPtr));
      SSL_SESSION_print (bio_stdout, SSL_get_session (sesolaptr->SslPtr));
   }
   BIO_printf (bio_stdout, "\n");

   BIO_printf (bio_stdout, "Client certificate:\n");
   PeerPtr = SSL_get_peer_certificate (sesolaptr->SslPtr);
   if (PeerPtr == NULL)
      BIO_puts (bio_stdout, "(no client certificate available)\n");
   else
   {
      X509_print (bio_stdout, PeerPtr);
      PEM_write_bio_X509 (bio_stdout, PeerPtr);
   }

   BIO_printf (bio_stdout, "\n----------\n");
}

#endif

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

BIO_METHOD *BIO_s_sesola()
{
   return (&sesola_method);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_new (BIO *bioptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_new()\n");

   bioptr->init = 1;
   bioptr->shutdown = 1;
   bioptr->num = 0;
   bioptr->flags = 0;
   /* the BIO pointer is set to the SESOLA structure by the calling routine */
   return (1);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_free (BIO *bioptr)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_free()\n");

   if (bioptr == NULL) return (0);
   /* don't free anything, the BIO pointer is set to the SESOLA structure */
   bioptr->ptr = NULL;
   return (1);
}
	
/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_read
(
BIO *bioptr,
char *DataPtr,
int DataSize
)
{
   register struct SeSoLaStruct  *sesolaptr;
   register struct RequestStruct  *rqptr;

   int  status,
        Count;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_read() %d %d\n", DataPtr, DataSize);

   sesolaptr = (struct SeSoLaStruct*)bioptr->ptr;
   rqptr = sesolaptr->rqptr;

   BIO_clear_retry_flags (bioptr);

   if (Debug)
      fprintf (stdout, "sesolaptr->ReadCount: %d\n", sesolaptr->ReadCount);

   if (sesolaptr->ReadCount)
   {
      /* will only be non-zero due to a previous non-blocking read */
      Count = sesolaptr->ReadCount;
      sesolaptr->ReadCount = 0;
#ifdef DBUG
      if (Debug && Count >= 0) SeSoLaDebugStream (DataPtr, Count);
#endif
      return (Count);
   }

   if (sesolaptr->ReadType == SESOLA_READ_DATA_BLOCK)
   {
      /* blocking I/O */
      sesolaptr->IoInProgress = false;
      status = NetReadRaw (rqptr, NULL, DataPtr, DataSize);
      /* if success return the number of characters written */
      if (VMSok (status)) return (rqptr->NetReadIOsb.Count);
      /* whoa, error */
      rqptr->NetReadIOsb.Status = status;
      return (-2);
   }
   else
   {
      /* non-blocking I/O */
      sesolaptr->IoInProgress = true;
      BIO_set_retry_read (bioptr);
      if (sesolaptr->ReadType == SESOLA_ACCEPT)
         sesolaptr->AcceptReadDataPtr = DataPtr;
      NetReadRaw (rqptr, &sesola_read_ast, DataPtr, DataSize);
      /* return indicating non-blocking I/O */
      return (-1);
   }
}

/*****************************************************************************/
/*
*/

int sesola_read_ast (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "sesola_read_ast() %%X%08.08X %d\n",
               rqptr->NetReadIOsb.Status, rqptr->NetReadIOsb.Count);

   /* the 'rqptr->SeSoLaPtr' is typeless! */
   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   sesolaptr->IoInProgress = false;

   if (VMSok (rqptr->NetReadIOsb.Status))
      sesolaptr->ReadCount = rqptr->NetReadIOsb.Count;
   else
      sesolaptr->ReadCount = -2;

   if (sesolaptr->ReadType == SESOLA_READ_DATA_NOBLOCK)
      SeSoLaRead (rqptr, sesolaptr->AstFunctionPtr,
                  sesolaptr->ReadDataPtr, sesolaptr->ReadDataSize);
   else
   if (sesolaptr->ReadType == SESOLA_ACCEPT)
      SeSoLaAccept (rqptr);
   else
   if (sesolaptr->ReadType == SESOLA_SHUTDOWN)
      SeSoLaEnd (rqptr);
   else
      ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_write
(
BIO *bioptr,
char *DataPtr,
int DataLength
)
{
   register struct SeSoLaStruct  *sesolaptr;
   register struct RequestStruct  *rqptr;

   int  status,
        Count;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_write() %d %d\n", DataPtr, DataLength);

   sesolaptr = (struct SeSoLaStruct*)bioptr->ptr;
   rqptr = sesolaptr->rqptr;

   BIO_clear_retry_flags (bioptr);

   if (Debug)
      fprintf (stdout, "sesolaptr->WriteCount: %d\n", sesolaptr->WriteCount);

   if (sesolaptr->WriteCount)
   {
      /* will only be non-zero due to a previous non-blocking write */
      if ((Count = sesolaptr->WriteCount) == -2) Count = 0;
      sesolaptr->WriteCount = 0;
#ifdef DBUG
      if (Debug) SeSoLaDebugStream (DataPtr, Count);
#endif
      /* writes always "succeed" :^( */
      return (DataLength);
   }

   if (sesolaptr->WriteType == SESOLA_WRITE_DATA_BLOCK)
   {
      /* blocking I/O */
      sesolaptr->IoInProgress = false;
      status = NetWriteRaw (sesolaptr->rqptr, NULL, DataPtr, DataLength);
      /* if success return the number of characters written */
      if (VMSok (status)) return (rqptr->NetWriteIOsb.Count);
      /* whoa, error */
      rqptr->NetWriteIOsb.Status = status;
      return (-2);
   }
   else
   {
      /* non-blocking I/O */
      sesolaptr->IoInProgress = true;
      /* indicate to the calling routine this write should be retried */
      BIO_set_retry_write (bioptr);
      NetWriteRaw (sesolaptr->rqptr, &sesola_write_ast, DataPtr, DataLength);
      /* return indicating non-blocking I/O */
      return (-1);
   }
}

/*****************************************************************************/
/*
*/

int sesola_write_ast (struct RequestStruct *rqptr)

{
   register struct SeSoLaStruct  *sesolaptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "sesola_write_ast() %%X%08.08X %d\n",
               rqptr->NetWriteIOsb.Status, rqptr->NetWriteIOsb.Count);

   sesolaptr = (struct SeSoLaStruct*)rqptr->SeSoLaPtr;

   sesolaptr->IoInProgress = false;

   if (VMSok (rqptr->NetWriteIOsb.Status))
      sesolaptr->WriteCount = rqptr->NetWriteIOsb.Count;
   else
      sesolaptr->WriteCount = -2;

   if (sesolaptr->WriteType == SESOLA_WRITE_DATA_NOBLOCK)
      SeSoLaWrite (rqptr, sesolaptr->AstFunctionPtr,
                   sesolaptr->WriteDataPtr, sesolaptr->WriteDataLength);
   else
   if (sesolaptr->WriteType == SESOLA_ACCEPT)
      SeSoLaAccept (rqptr);
   else
   if (sesolaptr->WriteType == SESOLA_SHUTDOWN)
      SeSoLaEnd (rqptr);
   else
      ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

long sesola_ctrl
(
BIO *bioptr,
int Command,
long Number,
char *Pointer
)
{
   int  value;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "sesola_ctrl() %d %d %d\n", Command, Number, Pointer);

   value = 1;
   switch (Command)
   {
      case BIO_CTRL_RESET:
         if (Debug) fprintf (stdout, "BIO_CTRL_RESET\n");
         return (1);

      case BIO_CTRL_EOF:
         if (Debug) fprintf (stdout, "BIO_CTRL_EOF\n");
         return (1);

      case BIO_CTRL_SET:
         if (Debug) fprintf (stdout, "BIO_CTRL_SET\n");
         bioptr->num = Number;
         return (1);

      case BIO_CTRL_SET_CLOSE:
         if (Debug) fprintf (stdout, "BIO_CTRL_SET_CLOSE\n");
         return (1);

      case BIO_CTRL_FLUSH:
         if (Debug) fprintf (stdout, "BIO_CTRL_FLUSH\n");
         return (1);

      case BIO_CTRL_DUP:
         if (Debug) fprintf (stdout, "BIO_CTRL_DUP\n");
         return (1);

      case BIO_CTRL_GET_CLOSE:
         if (Debug) fprintf (stdout, "BIO_CTRL_GET_CLOSE\n");
         return (0);

      case BIO_CTRL_INFO:
         if (Debug) fprintf (stdout, "BIO_CTRL_INFO\n");
         return (0);

      case BIO_CTRL_GET:
         if (Debug) fprintf (stdout, "BIO_CTRL_GET\n");
         return (0);

      case BIO_CTRL_PENDING:
         if (Debug) fprintf (stdout, "BIO_CTRL_PENDING\n");
         return (0);

      case BIO_CTRL_POP:
         if (Debug) fprintf (stdout, "BIO_CTRL_POP\n");
         return (0);

      case BIO_CTRL_PUSH:
         if (Debug) fprintf (stdout, "BIO_CTRL_PUSH\n");
         return (0);

      case BIO_CTRL_WPENDING:
         if (Debug) fprintf (stdout, "BIO_CTRL_WPENDING\n");
         return (0);

      default:
         if (Debug) fprintf (stdout, "default:\n");
         return (0);
   }
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_gets
(
BIO *bioptr,
char *StringPtr,
int StringSize
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_gets()\n");
   return(0);
}

/*****************************************************************************/
/*
Implements a required function of a OpenSSL BIO_METHOD.
*/

int sesola_puts
(
BIO *bioptr,
char *StringPtr
)
{
    int  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "sesola_puts() |%s|\n", StringPtr);

    if (StringPtr == NULL) return (0);
    Length = strlen (StringPtr);
    if (Length == 0) return (0);
    return (sesola_write (bioptr, StringPtr, Length));
}

/*****************************************************************************/
/*
For compilations without SSL these functions provide LINKage stubs for the
rest of the HTTPd modules, allowing for just recompiling the SESOLA module to
integrate the SSL functionality.
*/

/*********************/
#else  /* not SESOLA */
/*********************/

#ifdef DBUG
extern boolean Debug;
#else
#define Debug 0 
#endif

/* global storage */
boolean  ProtocolHttpsAvailable,
         ProtocolHttpsConfigured;
/* i.e. disabled */
int  SeSoLaProtocolVersion = -1;
char  HttpdSesola [] = "";

/* external storage */
extern char  ErrorSanityCheck[];
extern char  Utility[];

SeSoLaInit ()
{
   ProtocolHttpsAvailable = false;
   if (SeSoLaProtocolVersion <= 0) return;
   fprintf (stdout, "%%%s-E-SSL, non-SSL version\n", Utility);
   exit (STS$K_ERROR | STS$M_INHIB_MSG);
}

SeSoLaInitServiceList ()
{
   return;
}

SeSoLaBegin (struct RequestStruct *rqptr)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaAccept (struct RequestStruct *rqptr)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaEnd (struct RequestStruct *rqptr)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaFree (struct RequestStruct *rqptr)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaShutdown (struct RequestStruct *rqptr)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

int SeSoLaWrite
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataLength
)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaRead
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *DataPtr,
int DataSize
)
{
   ErrorExitVmsStatus (0, ErrorSanityCheck, FI_LI);
}

SeSoLaReport
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   ErrorGeneral (rqptr, "This is a non-SSL version.", FI_LI);
   SysDclAst (NextTaskFunction, rqptr);
}

/************************/
#endif  /* ifdef SESOLA */
/************************/

/*****************************************************************************/

